@@ -1,6 +1,6 @@
 /* 
  * Copyright (C) 2006, 2007, 2008 OpenedHand Ltd.
- * Copyright (C) 2009 Nokia Corporation, all rights reserved.
+ * Copyright (C) 2009 Nokia Corporation.
  *
  * Author: Jorn Baayen <jorn@openedhand.com>
  *         Zeeshan Ali (Khattak) <zeeshanak@gnome.org>
@@ -72,9 +72,17 @@ typedef unsigned long in_addr_t;
 /* Size of the buffer used for reading from the socket */
 #define BUF_SIZE 1024
 
-G_DEFINE_TYPE (GSSDPClient,
-               gssdp_client,
-               G_TYPE_OBJECT);
+static void
+gssdp_client_initable_iface_init (gpointer g_iface,
+                                  gpointer iface_data);
+
+G_DEFINE_TYPE_EXTENDED (GSSDPClient,
+                        gssdp_client,
+                        G_TYPE_OBJECT,
+                        0,
+                        G_IMPLEMENT_INTERFACE
+                                (G_TYPE_INITABLE,
+                                 gssdp_client_initable_iface_init));
 
 struct _GSSDPClientPrivate {
         GMainContext      *main_context;
@@ -84,10 +92,9 @@ struct _GSSDPClientPrivate {
         char              *host_ip;
         char              *network;
 
-        GError            **error;
-
         GSSDPSocketSource *request_socket;
         GSSDPSocketSource *multicast_socket;
+        GSSDPSocketSource *search_socket;
 
         gboolean           active;
 };
@@ -99,8 +106,7 @@ enum {
         PROP_IFACE,
         PROP_NETWORK,
         PROP_HOST_IP,
-        PROP_ACTIVE,
-        PROP_ERROR
+        PROP_ACTIVE
 };
 
 enum {
@@ -125,7 +131,18 @@ multicast_socket_source_cb    (GIOChannel   *source,
                                GIOCondition  condition,
                                gpointer      user_data);
 static gboolean
-init_network_info             (GSSDPClient  *client);
+search_socket_source_cb       (GIOChannel   *source,
+                               GIOCondition  condition,
+                               gpointer      user_data);
+
+static gboolean
+init_network_info             (GSSDPClient  *client,
+                               GError      **error);
+
+static gboolean
+gssdp_client_initable_init    (GInitable     *initable,
+                               GCancellable  *cancellable,
+                               GError       **error);
 
 static void
 gssdp_client_init (GSSDPClient *client)
@@ -142,33 +159,45 @@ gssdp_client_init (GSSDPClient *client)
 }
 
 static void
-gssdp_client_constructed (GObject *object)
+gssdp_client_initable_iface_init (gpointer g_iface,
+                                  gpointer iface_data)
 {
-        GSSDPClient *client = GSSDP_CLIENT (object);
-        GError *error = NULL;
+        GInitableIface *iface = (GInitableIface *)g_iface;
+        iface->init = gssdp_client_initable_init;
+}
+
+static gboolean
+gssdp_client_initable_init (GInitable     *initable,
+                            GCancellable  *cancellable,
+                            GError       **error)
+{
+        GSSDPClient *client = GSSDP_CLIENT (initable);
+        GError *internal_error = NULL;
 #ifdef G_OS_WIN32
         WSADATA wsaData = {0};
         if (WSAStartup (MAKEWORD (2,2), &wsaData) != 0) {
                 gchar *message;
 
                 message = g_win32_error_message (WSAGetLastError ());
-                g_set_error_literal (client->priv->error,
+                g_set_error_literal (error,
                                      GSSDP_ERROR,
                                      GSSDP_ERROR_FAILED,
                                      message);
                 g_free (message);
+
+                return FALSE;
         }
 #endif
 
         /* Make sure all network info is available to us */
-        if (!init_network_info (client))
-                return;
+        if (!init_network_info (client, &internal_error))
+                goto errors;
 
         /* Set up sockets (Will set errno if it failed) */
         client->priv->request_socket =
                 gssdp_socket_source_new (GSSDP_SOCKET_SOURCE_TYPE_REQUEST,
                                          gssdp_client_get_host_ip (client),
-                                         &error);
+                                         &internal_error);
         if (client->priv->request_socket != NULL) {
                 gssdp_socket_source_set_callback
                         (client->priv->request_socket,
@@ -181,20 +210,53 @@ gssdp_client_constructed (GObject *object)
         client->priv->multicast_socket =
                 gssdp_socket_source_new (GSSDP_SOCKET_SOURCE_TYPE_MULTICAST,
                                          gssdp_client_get_host_ip (client),
-                                         &error);
+                                         &internal_error);
         if (client->priv->multicast_socket != NULL) {
                 gssdp_socket_source_set_callback
                         (client->priv->multicast_socket,
                          (GSourceFunc) multicast_socket_source_cb,
                          client);
+        } else {
+                goto errors;
         }
 
+        /* Setup send socket. For security reasons, it is not recommended to
+         * send M-SEARCH with source port == SSDP_PORT */
+        client->priv->search_socket = gssdp_socket_source_new
+                                        (GSSDP_SOCKET_SOURCE_TYPE_SEARCH,
+                                         gssdp_client_get_host_ip (client),
+                                         &internal_error);
+        if (client->priv->search_socket != NULL) {
+                gssdp_socket_source_set_callback
+                                        (client->priv->search_socket,
+                                         (GSourceFunc) search_socket_source_cb,
+                                         client);
+        }
  errors:
-        if (!client->priv->request_socket || !client->priv->multicast_socket) {
-                if (client->priv->error) {
-                        g_propagate_error (client->priv->error, error);
+        if (!client->priv->request_socket ||
+            !client->priv->multicast_socket ||
+            !client->priv->search_socket) {
+                g_propagate_error (error, internal_error);
+
+                if (client->priv->request_socket) {
+                        g_object_unref (client->priv->request_socket);
+
+                        client->priv->request_socket = NULL;
                 }
-                return;
+
+                if (client->priv->multicast_socket) {
+                        g_object_unref (client->priv->multicast_socket);
+
+                        client->priv->multicast_socket = NULL;
+                }
+
+                if (client->priv->search_socket) {
+                        g_object_unref (client->priv->search_socket);
+
+                        client->priv->search_socket = NULL;
+                }
+
+                return FALSE;
         }
 
         gssdp_socket_source_attach (client->priv->request_socket,
@@ -202,6 +264,11 @@ gssdp_client_constructed (GObject *object)
 
         gssdp_socket_source_attach (client->priv->multicast_socket,
                                     client->priv->main_context);
+
+        gssdp_socket_source_attach (client->priv->search_socket,
+                                    client->priv->main_context);
+
+        return TRUE;
 }
 
 static void
@@ -266,9 +333,6 @@ gssdp_client_set_property (GObject      *object,
                 gssdp_client_set_main_context (client,
                                                g_value_get_pointer (value));
                 break;
-        case PROP_ERROR:
-                client->priv->error = g_value_get_pointer (value);
-                break;
         case PROP_IFACE:
                 client->priv->iface = g_value_dup_string (value);
                 break;
@@ -302,6 +366,11 @@ gssdp_client_dispose (GObject *object)
                 client->priv->multicast_socket = NULL;
         }
 
+        if (client->priv->search_socket) {
+                g_object_unref (client->priv->search_socket);
+                client->priv->search_socket = NULL;
+        }
+
         /* Unref the context */
         if (client->priv->main_context) {
                 g_main_context_unref (client->priv->main_context);
@@ -336,7 +405,6 @@ gssdp_client_class_init (GSSDPClientClass *klass)
 
 	object_class = G_OBJECT_CLASS (klass);
 
-	object_class->constructed = gssdp_client_constructed;
 	object_class->set_property = gssdp_client_set_property;
 	object_class->get_property = gssdp_client_get_property;
 	object_class->dispose      = gssdp_client_dispose;
@@ -377,25 +445,6 @@ gssdp_client_class_init (GSSDPClientClass *klass)
                           G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
                           G_PARAM_STATIC_BLURB));
 
-        /**
-         * GSSDPClient:error
-         *
-         * Internal property.
-         *
-         * Stability: Private
-         **/
-        g_object_class_install_property
-                (object_class,
-                 PROP_ERROR,
-                 g_param_spec_pointer
-                         ("error",
-                          "Error",
-                          "Location where to store the constructor GError, "
-                          "if any.",
-                          G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY |
-                          G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
-                          G_PARAM_STATIC_BLURB));
-
         /**
          * GSSDPClient:interface
          *
@@ -504,7 +553,7 @@ gssdp_client_class_init (GSSDPClientClass *klass)
 /**
  * gssdp_client_new
  * @main_context: The #GMainContext to associate with, or NULL
- * @interface: The name of the network interface, or %NULL for auto-detection.
+ * @iface: The name of the network interface, or %NULL for auto-detection.
  * @error: Location to store error, or NULL
  *
  * Return value: A new #GSSDPClient object.
@@ -514,11 +563,12 @@ gssdp_client_new (GMainContext *main_context,
                   const char   *iface,
                   GError      **error)
 {
-        return g_object_new (GSSDP_TYPE_CLIENT,
-                             "main-context", main_context,
-                             "interface", iface,
-                             "error", error,
-                             NULL);
+        return g_initable_new (GSSDP_TYPE_CLIENT,
+                               NULL,
+                               error,
+                               "main-context", main_context,
+                               "interface", iface,
+                               NULL);
 }
 
 /**
@@ -541,7 +591,7 @@ gssdp_client_set_main_context (GSSDPClient  *client,
  * gssdp_client_get_main_context
  * @client: A #GSSDPClient
  *
- * Return value: The #GMainContext @client is associated with, or NULL.
+ * Returns: (transfer none): The #GMainContext @client is associated with, or NULL.
  **/
 GMainContext *
 gssdp_client_get_main_context (GSSDPClient *client)
@@ -685,16 +735,17 @@ gssdp_client_get_active (GSSDPClient *client)
  * Sends @message to @dest_ip.
  **/
 void
-_gssdp_client_send_message (GSSDPClient *client,
-                            const char  *dest_ip,
-                            gushort      dest_port,
-                            const char  *message)
+_gssdp_client_send_message (GSSDPClient      *client,
+                            const char       *dest_ip,
+                            gushort           dest_port,
+                            const char       *message,
+                            _GSSDPMessageType type)
 {
         gssize res;
         GError *error = NULL;
-        GSocket *socket = NULL;
         GInetAddress *inet_address = NULL;
         GSocketAddress *address = NULL;
+        GSocket *socket;
 
         g_return_if_fail (GSSDP_IS_CLIENT (client));
         g_return_if_fail (message != NULL);
@@ -711,9 +762,16 @@ _gssdp_client_send_message (GSSDPClient *client,
         if (dest_port == 0)
                 dest_port = SSDP_PORT;
 
+        if (type == _GSSDP_DISCOVERY_REQUEST)
+                socket = gssdp_socket_source_get_socket
+                                        (client->priv->search_socket);
+        else
+                socket = gssdp_socket_source_get_socket
+                                        (client->priv->request_socket);
+
         inet_address = g_inet_address_new_from_string (dest_ip);
         address = g_inet_socket_address_new (inet_address, dest_port);
-        socket = gssdp_socket_source_get_socket (client->priv->request_socket);
+
         res = g_socket_send_to (socket,
                                 address,
                                 message,
@@ -768,7 +826,9 @@ parse_http_request (char                *buf,
                     SoupMessageHeaders **headers,
                     int                 *type)
 {
-        char *req_method;
+        char *req_method = NULL;
+        char *path = NULL;
+        SoupHTTPVersion version;
 
         *headers = soup_message_headers_new (SOUP_MESSAGE_HEADERS_REQUEST);
 
@@ -776,8 +836,10 @@ parse_http_request (char                *buf,
                                         len,
                                         *headers,
                                         &req_method,
-                                        NULL,
-                                        NULL) == SOUP_STATUS_OK) {
+                                        &path,
+                                        &version) == SOUP_STATUS_OK &&
+            version == SOUP_HTTP_1_1 &&
+            (path && g_ascii_strncasecmp (path, "*", 1) == 0)) {
                 if (g_ascii_strncasecmp (req_method,
                                          SSDP_SEARCH_METHOD,
                                          strlen (SSDP_SEARCH_METHOD)) == 0)
@@ -796,6 +858,12 @@ parse_http_request (char                *buf,
                 soup_message_headers_free (*headers);
                 *headers = NULL;
 
+                if (path)
+                        g_free (path);
+
+                if (req_method)
+                        g_free (req_method);
+
                 return FALSE;
         }
 }
@@ -997,6 +1065,18 @@ multicast_socket_source_cb (GIOChannel  *source,
         return socket_source_cb (client->priv->multicast_socket, client);
 }
 
+static gboolean
+search_socket_source_cb (GIOChannel  *source,
+                         GIOCondition condition,
+                         gpointer     user_data)
+{
+        GSSDPClient *client;
+
+        client = GSSDP_CLIENT (user_data);
+
+        return socket_source_cb (client->priv->search_socket, client);
+}
+
 #ifdef G_OS_WIN32
 static gboolean
 is_primary_adapter (PIP_ADAPTER_ADDRESSES adapter)
@@ -1241,7 +1321,7 @@ get_host_ip (char **iface, char **network)
 }
 
 static gboolean
-init_network_info (GSSDPClient *client)
+init_network_info (GSSDPClient *client, GError **error)
 {
         gboolean ret = TRUE;
 
@@ -1251,16 +1331,14 @@ init_network_info (GSSDPClient *client)
                                      &client->priv->network);
 
         if (client->priv->iface == NULL) {
-                if (client->priv->error)
-                        g_set_error (client->priv->error,
+                g_set_error_literal (error,
                                      GSSDP_ERROR,
                                      GSSDP_ERROR_FAILED,
                                      "No default route?");
 
                 ret = FALSE;
         } else if (client->priv->host_ip == NULL) {
-                if (client->priv->error)
-                        g_set_error (client->priv->error,
+                        g_set_error (error,
                                      GSSDP_ERROR,
                                      GSSDP_ERROR_NO_IP_ADDRESS,
                                      "Failed to find IP of interface %s",
