@@ -740,7 +740,7 @@ message_received_cb (GSSDPClient        *client,
                      gpointer            user_data)
 {
         GSSDPResourceGroup *resource_group;
-        const char *target, *mx_str, *version_str;
+        const char *target, *mx_str, *version_str, *man;
         gboolean want_all;
         int mx, version;
         GList *l;
@@ -768,12 +768,19 @@ message_received_cb (GSSDPClient        *client,
 
         /* Extract MX */
         mx_str = soup_message_headers_get_one (headers, "MX");
-        if (!mx_str || atoi (mx_str) == 0) {
+        if (!mx_str || atoi (mx_str) <= 0) {
                 g_warning ("Discovery request did not have a valid MX header");
  
                 return;
         }
- 
+
+        man = soup_message_headers_get_one (headers, "MAN");
+        if (!man) {
+                g_warning ("Discovery request did not have a valid MAN header");
+
+                return;
+        }
+
         mx = atoi (mx_str);
 
         /* Extract version */
@@ -915,7 +922,8 @@ discovery_response_timeout (gpointer user_data)
         _gssdp_client_send_message (client,
                                     response->dest_ip,
                                     response->dest_port,
-                                    message);
+                                    message,
+                                    _GSSDP_DISCOVERY_RESPONSE);
 
         g_free (message);
         g_free (date_str);
@@ -970,7 +978,8 @@ process_queue (gpointer data)
                 _gssdp_client_send_message (client,
                                             NULL,
                                             0,
-                                            message);
+                                            message,
+                                            _GSSDP_DISCOVERY_RESPONSE);
                 g_free (message);
 
                 return TRUE;
