@@ -1,6 +1,6 @@
 /* 
  * Copyright (C) 2006, 2007, 2008 OpenedHand Ltd.
- * Copyright (C) 2009 Nokia Corporation, all rights reserved.
+ * Copyright (C) 2009 Nokia Corporation.
  * Copyright (C) 2010 Jens Georg <mail@jensge.org>
  *
  * Author: Jorn Baayen <jorn@openedhand.com>
@@ -205,6 +205,7 @@ gssdp_socket_source_do_init (GInitable     *initable,
                                             inner_error,
                                             "Failed to set TTL");
 
+                goto error;
         }
         /* Set up additional things according to the type of socket desired */
         if (self->priv->type == GSSDP_SOCKET_SOURCE_TYPE_MULTICAST) {
@@ -239,15 +240,19 @@ gssdp_socket_source_do_init (GInitable     *initable,
                                                           SSDP_PORT);
 #endif
         } else {
+                guint port = SSDP_PORT;
+
+                /* Chose random port For the socket source used by M-SEARCH */
+                if (self->priv->type == GSSDP_SOCKET_SOURCE_TYPE_SEARCH)
+                        port = 0;
                 bind_address = g_inet_socket_address_new (iface_address,
-                                                          SSDP_PORT);
+                                                          port);
         }
 
-#ifdef G_OS_WIN32
-        /* normally g_socket_bind does this, but it is disabled on
+        /* Normally g_socket_bind does this, but it is disabled on
          * windows since SO_REUSEADDR has different semantics
-         * there. Nevertheless, there's no way without for
-         * multicast sockets
+         * there, also we nees SO_REUSEPORT on OpenBSD. This is a nop
+         * everywhere else.
          */
         if (!gssdp_socket_reuse_address (self->priv->socket,
                                          TRUE,
@@ -259,7 +264,7 @@ gssdp_socket_source_do_init (GInitable     *initable,
 
                 goto error;
         }
-#endif
+
         /* Bind to requested port and address */
         if (!g_socket_bind (self->priv->socket,
                             bind_address,
@@ -400,7 +405,7 @@ gssdp_socket_source_class_init (GSSDPSocketSourceClass *klass)
                          "Type",
                          "Type of socket-source (Multicast/Unicast)",
                          GSSDP_SOCKET_SOURCE_TYPE_REQUEST,
-                         GSSDP_SOCKET_SOURCE_TYPE_MULTICAST,
+                         GSSDP_SOCKET_SOURCE_TYPE_SEARCH,
                          GSSDP_SOCKET_SOURCE_TYPE_REQUEST,
                          G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY |
                          G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
